/* background routines */

#define NUM_SCAN_LINES 1280
#include "graph.h"
#include "pallette.h"
#include "poly.h"
#include <dos.h>
#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>

struct SCANLINE {
   char pri;       /* primary color */
   char sec;       /* secondary color */
   int start;     /* scan dot of start of 2nd color */
   int end;       /* scan dot of end of 2nd color, this dot not drawn */
   };

/* the scanlines wrap at 2560 pixels, but the details can't cross zero */
/* scanline[] defines 1280 scanlines each with a possible 2560 pixels */
/* this holds background info for 360 deg pans and 180 deg altitude changes */

/* added 2nd scanline array so can project the runway into these arrays */

static struct SCANLINE scanline[NUM_SCAN_LINES];
static struct SCANLINE scanline2[NUM_SCAN_LINES];

static int startline= NUM_SCAN_LINES-1;    /* top scanline on screen */
static int startpixel= 0;

/* for runway */
static struct POINT runpoints[6];
static struct POLY runway;
int runverts[4]= {0,1,2,3};
struct SCANLINE  *run, *clear;

addrunpoint(int col, int row){

   if( row < 0 || row > NUM_SCAN_LINES ) return;  /* array bounds */
  /* add runway color */
  /* swap points if necessary so start is lower pixel value */
   if( run[row].sec == 0 ){
      run[row].sec= RUNWAY;
      run[row].start= col;
      run[row].end= -1;
      }
   else{
      if( run[row].end != -1 ) return;
      if( run[row].start < col ) run[row].end= col;
      else{
         run[row].end= run[row].start;
         run[row].start= col;
         }
      }
}


void draw_runway(struct POINT * p1, int count){
struct POINT *p2, *p3;


   if( count < 3 ) return;

   p2= p3= p1;
   while( --count ){
      traceline( p2++, ++p3, &addrunpoint );
      }
   traceline( p2, p1, &addrunpoint );

}



/* from Abrash with changes for my structures */
/* clip polygons */
int clip( int count, struct POINT *in, struct POINT *out,
          struct POINT *line, double distance){
struct POINT *newin,*firstin;
int i,curin, nextin;
double curdot, nextdot,scale;
int points= 0;

curdot= in->x*line->x + in->y*line->y;
curin= ( curdot >=distance );
firstin= in;

for( i= 0; i < count; ++i ){

   newin= ( i < count-1 ) ? in+1 : firstin;
   if( curin ){          /* add point if inside the view area */
      out->x = in->x;
      out->y = in->y;
      out->z = in->z;
      ++points;
      ++out;
      }
   nextdot= newin->x*line->x + newin->y*line->y;
   nextin= (nextdot >= distance );

/* add a point if next point crosses our clip plane */
   if( curin != nextin ){
      scale= ( distance - curdot ) / (nextdot - curdot );
      out->x = in->x + (( newin->x - in->x ) * scale );
      out->y = in->y + (( newin->y - in->y ) * scale );
      out->z = in->z + (( newin->z - in->z ) * scale );
      ++points;
      ++out;
      }
   curdot= nextdot;
   curin= nextin;
   ++in;
   }

return points;
}



/* project the 4 runway points to row and col values */
void project_runway(int newpix){
int i;
struct POINT dest[10],dest2[10];
extern struct POINT view_position;
extern double view[3][3];
int count;
double dist;
struct POINT clip1= {     /* line x = y */
   0.707, 0.707, 0.0, 0, 0, 0.0
   };
struct POINT clip2= {     /* line  x = -y */
  -0.707, 0.707, 0.0, 0, 0, 0.0
   };

/* translate to view space and project points in view space */
   for( i= 0; i < 4; ++i ){
      dest[i].x= runpoints[i].x - view_position.x;
      dest[i].y= runpoints[i].y - view_position.y;
      dest[i].z= runpoints[i].z - view_position.z;
      xformpoint(view,&dest[i],&dest2[i]);
      }

/* clip to view */
   count= 4;
   count= clip(count, &dest2[0], &dest[0] , &clip1, 1.0);
   count= clip(count, &dest[0], &dest2[0], &clip2 , 1.0);


   for( i= 0; i < count; ++i ){
      dist= 0.0015 * sqrt( dest2[i].x*dest2[i].x + dest2[i].y*dest2[i].y +
            dest2[i].z*dest2[i].z );    /* was 0.0025 */

      dest2[i].row=640 -  407*atan(10.0/dest2[i].y); /* works ok*/
      dest2[i].col= newpix + 320 + dest2[i].x / dist;
      }

   draw_runway( &dest2[0], count );


}


/* draw just the secondary scan, but in the passed color */
/* we can draw or erase details depending on color passed */
void drawsec(struct SCANLINE *new,int pixel,int row,int color){
int start, end;

   /* see if in bounds */
   if( pixel > new->end ) pixel-= 2560;

   if(  (start= new->start - pixel) < 0 ) start= 0;
   if(  (end= new->end - pixel) > 640 ) end = 640;
   if( start > end ) return;

   mypencolor( color );
   hline3(start,row,end-start);

}

/* try to draw whats needed of overlapping lines */
void mergeline(struct SCANLINE *new, struct SCANLINE *old,
               int newpixel, int oldpixel, int row){

int newstart, oldstart;
int newend, oldend;
int start, end;

   if( newpixel > new->end )  newpixel-= 2560;
   if( oldpixel > old->end )  oldpixel-= 2560;

      /* find the bounds of the lines */
   if( (newstart= new->start - newpixel) < 0 ) newstart= 0;
   if( (newend= new->end - newpixel) > 640 ) newend= 640;
   if( (oldstart= old->start - oldpixel) < 0 ) oldstart= 0;
   if( (oldend= old->end - oldpixel) > 640 ) oldend= 640;

   /* have a number of possibilities */
       /* nothing on screen */
   if( newstart > newend && oldstart > oldend ) return;

       /* just one on screen */
   if( newstart > newend ) {          /* erase old line */
      mypencolor(new->pri);
      hline3(oldstart,row,oldend-oldstart);
      return;
      }
   if( oldstart > oldend ){           /* just draw new */
      mypencolor(new->sec);
      hline3(newstart,row,newend-newstart);
      return;
      }

       /* we now know both on screen */
       /* should only have to draw two lines,
                     question is what color and where(thats 2 questions) */
   if( newstart < oldstart ){       /* draw new segment */
      mypencolor(new->sec);
      start= newstart;
      end=  (newend < oldstart ) ? newend: oldstart;
      }
   else{                       /* erase old segment */
      mypencolor(new->pri);
      start= oldstart;
      end= ( oldend < newstart ) ? oldend: newstart;
      }
   hline3(start, row, end-start);

   if( newend > oldend ){      /* draw new segment */
      mypencolor(new->sec);
      end= newend;
      start= (newstart > oldend) ? newstart:oldend;
      }
   else{                       /* erase old segment */
      mypencolor(new->pri);
      end= oldend;
      start= ( oldstart > newend ) ? oldstart:newend;
      }
   hline3(start, row, end-start);



}

/* draw complete scanline at this row */
void wholeline(struct SCANLINE *new,int pixel,int row){
int num;
int i;
int j;


    if( new->sec == 0 ){      /* no details */
       mypencolor(new->pri);  /* we may come here if compared line has detail*/

       hline3(0,row,640);
       return;
       }

    num= 0;
    while( num < 640 ){
       if( pixel >= new->start && pixel < new->end ){
          mypencolor(new->sec);
          i= new->end - pixel;       /* we are in the secondary color */
          j= 640 - i - num;
          if( j < 0 ) i= i +j;
          hline3(num,row,i);
          pixel+= i;
          num+= i;
          }
       else if( pixel >= new->end ){      /* end section of scanline */
          mypencolor(new->pri);
          i= 2560 - pixel;
          j= 640 - i - num;
          if( j < 0 ) i= i + j;
          hline3(num,row,i);
          pixel= 0;
          num+= i;
          }
       else{       /* 1st section of the line */
          mypencolor(new->pri);
          i= new->start - pixel;
          j= 640 - i - num;
          if( j < 0 ) i= i + j;
          hline3(num,row,i);
          pixel+= i;
          num+= i;
          }

       }

}

/* draw the background changes */
void draw_back( int newline, int newpixel ){
struct SCANLINE *old, *new;
int i;
static int flip;


   if( newline == startline && newpixel == startpixel ) return;

/* alternate scanline arrays */
   if( flip ){
      old= &scanline[startline];
      new= &scanline2[newline];
      run= scanline2;
      clear= scanline;
      }
   else{
      old= &scanline2[startline];
      new= &scanline[newline];
      run= scanline;
      clear= scanline2;
      }
   flip ^= 1;

/* project and scan runway points into new scanline array */
   project_runway(newpixel);

   planemask(0xc);   /* draw only to background planes */
   for( i= 0; i < 480; ++i,--old,--new ){
    /* if pri colors are different, draw complete scan line */
         if( new->pri != old->pri ){
            wholeline(new,newpixel,i);
            continue;
            }
    /* have same pri color, if no sec then done */
         if( new->sec == 0 && old->sec == 0 ) continue;

    /* test for both lines having sec color */
         if( new->sec != 0 && old->sec != 0 ){
            if( new->sec == old->sec )  /* same color ?? */
                        mergeline(new,old,newpixel,startpixel,i);
            else wholeline(new,newpixel,i); /* different sec color */
            continue;
            }

     /* we now know pri are same and only one has sec color */
         if( new->sec ) drawsec(new,newpixel,i,new->sec);
     /* else erase old sec with pri color */
         else  drawsec(old,startpixel,i,new->pri);

         } /* end of 480 lines */

/* would be good if could get this inside the main background loop */
  for( i= 0; i < 641; ++i )  if ( clear[i].sec == RUNWAY ) clear[i].sec= 0;


   startline= newline;   startpixel= newpixel;
   /* maybe should reset planemask, maybe not */
}



/* read the backgnd.txt file for details of sceen */
void read_back(){
FILE *fp;
int vmult,hmult;
int gpix;
int start, end;
int color;
int line;
int data;

int i;
char temp[256];

   fp= fopen("backgnd.dat","r");
   if( fp == 0 ) return;

   while( fscanf(fp,"%d",&line) != EOF ){
      fgets(temp,256,fp);  /* eat any comments */
      fscanf(fp,"%d",&data);
      fgets(temp,256,fp);  /* eat any comments */
      fscanf(fp,"%d",&vmult);
      fgets(temp,256,fp);  /* eat any comments */
      fscanf(fp,"%d",&gpix);
      fgets(temp,256,fp);  /* eat any comments */
      fscanf(fp,"%d",&hmult);
      fgets(temp,256,fp);  /* eat any comments */
      fscanf(fp,"%d",&color);
      fgets(temp,256,fp);  /* eat any comments */

      while( data-- ){
         fscanf(fp,"%d%d",&start,&end);
         start= start*hmult + gpix;
         end= end*hmult + gpix;
         for( i= 0; i < vmult; ++i ){
            scanline2[line].sec= scanline[line].sec= color;
            scanline2[line].start= scanline[line].start= start;
            scanline2[line].end= scanline[line].end= end;
            --line;
            }
         }
      }

   fclose(fp);
}


void init_back(){   /* initialize the background */
int i;

/* set up runway points and poly */
   for( i= 0; i < 4; ++i ){
      runpoints[i].z= -10;
      }
    runpoints[0].x= -100;
    runpoints[0].y= 30;
    runpoints[1].x= 100;
    runpoints[1].y= 30;
    runpoints[2].x= 100;
    runpoints[2].y= 70;
    runpoints[3].x= -100;
    runpoints[3].y= 70;

   runway.color= RUNWAY;
   runway.no_points= 6;
   runway.verts= &runverts[0];

/* set up ground and sky */
   for( i= 0; i < 500; ++i ){
      scanline2[i].pri= scanline[i].pri= GRASS;
      }
   for( ; i < 641;  ) scanline2[i].pri= scanline[i++].pri= GRASS;
   for(; i < NUM_SCAN_LINES; ) scanline2[i].pri= scanline[i++].pri= SKY; /* defaults as */

   read_back();
   draw_back( NUM_SCAN_LINES/2+480/2, 0);

}


