
/*
                            FlyRc


  copyright (c) 1998  by Ron Carr as unpublished work, all rights reserved.
  This code can not be used for commercial purpose.

  Many algorithms are just a trial of some ideas I had, and probably not
  the best way to produce the desired result.  This is 640x480 vga non page
  flipped mode.
*/

#include <stdio.h>
#include "graph.h"
#include "vga.h"
#include <time.h>
#include "pallette.h"
#include "poly.h"
#include <math.h>
#include <string.h>
#include <conio.h>
#include <dos.h>     /* for int86 */
#include <direct.h>  /* for finding plane choices */
#include <ctype.h>

struct JOY {
   int ax;     /* x value */
   int ay;     /* y value */
   int bx;     /* 2nd joystick x and y */
   int by;
   int fa1;    /* fire button 1 */
   int fa2;    /* fire button 2 */
   int fb1;    /* 2nd joystick fire buttons */
   int fb2;
   };

struct JOY joy;
void draw_front();
void project( struct POINT *p, int shadow_flag);



#define UP 72
#define DOWN 80
#define LEFT 75
#define RIGHT 77
/* #define PAGEUP 73 */
/* #define PAGEDOWN 81 */


/* stack pallette for P1 P2 and shadows in forground */
int mypal[16]={SKYC,SKYC,P1C,P2C,GRASSC,SHADOW_GRASSC,P1C,P2C,
             RUNWAYC,SHADOW_RUNWAYC,P1C,P2C,CLOUDC,CLOUDC,P1C,P2C};

int p1c= P1C;    /* allow to change plane colors */
int p2c= P2C;

/* global vars for our one(1) object - a plane */
#define MAX_POINT 200
#define MAX_POLY 200
#define POINTSINPOLY 30   /* limit to vertices of a poly - used in shadows */

/*                     flight model
   lift and gravity balance so they are ignored
   plane has rotation toward ground that must be countered by elevator
   when stalled the plane drops and  rotation toward ground increases
   gravity assists or resists thrust in the Z direction only
   drag is a function of speed squared

   low speed model has been reworked
*/


/* variables needed for flight */
double thrust;    /* engine power */
int etrim;       /* elevator key step */
int atrim;       /* airelon key steps */
int ejoy;        /* joystick inputs */
int ajoy;
int ejoycal;
int ajoycal;
double keyrate;

/* for tx transmitter input */
int status;     /* lpt1 ports */
int control;
int invert;
int nominal;      /* typical count for centered stick */
int rcsticks= 0;   /* flag for tx trans present */
int iiface;        /* what input device is being used */
int av,ev,tv,rv;   /* assignment of channels via config file */

struct POINT position;         /* planes position */
struct POINT view_position;    /* starts at zero */
double speed = 1.0;            /* speed and xform == velocity vector */
                        /* speed is 60 ft/sec,if have max frame rate */
/* these will be contants that we can vary */
double power= 1;
double drag= 1;
double lowdrag= 1;       /* low speed drag */
double inertia= 1;
double gtrim= 1;       /* trim torque due to gravity */
double ztrim= 1;       /* turn due to wings not level */
double gthrust= 1.0;     /* thrust due to gravity */
double sdrop= 2;  /* plane drops when stalled, zero for flyby */
double stall= 10.0;   /* stall speed */
double best_speed;
double drag_floor;
double gstall;
double zstall;


int nojoy= 1;
double ejoyspeed;       /* fudge factors for input devices */
double ajoyspeed;
double eradiospeed;
double aradiospeed;
double espeed;
double aspeed;
double rspeed;     /* rudder */
int rudder;

int no_points;           /* actual number of points */
int no_polys;            /* actual number of polygons */

struct POINT points[MAX_POINT];    /* simple list of points */
struct POLY polys[MAX_POLY];       /* simple list of poly's */

double xform[3][3]={        /* plane xform */
   1.0,  0.0,  0.0,
   0.0,  1.0,  0.0,
   0.0,  0.0,  1.0 };

double view[3][3]={         /* viewer xform */
   1.0,  0.0,  0.0,
   0.0,  1.0,  0.0,
   0.0,  0.0,  1.0 };


int   pix= 0;      /* background starting pixel */
int backline= 0;  /* 880;   /* background starting line */
int row_offset= 380;    /* sliding window for offsceen image */
double viewangle;      /* updown view angle */
double frames= 20.0;   /* expected max frame rate is 60, make this low for raw speed indication */


/*
  our view coordinate system is defined so that the x axis is in line with our
  shoulders with pos values to the right.  The y axis is in line with
  our line of sight with pos value going away
  The z axis determines the altitude.

This comment no longer applys - we now have a view xform that determines
the orientation of the viewpoint ( the rotating system worked fine though )
  The system is rotated with the object as it moves so that:
    the x value of position will always be zero.
    the y value should always be positive
*/

extern void AppendRotationX( double * xform, double angle );
extern void appendx( double * xform, double angle );
extern void appendy( double * xform, double angle );
extern void appendz( double * xform, double angle );
extern void dumpscan( );
extern void drawpoly( struct POLY *p1, struct POINT *p2 );
extern void xformpoint( double *xform, struct POINT *src, struct POINT *dest );
extern void draw_back( int scan, int pixel );
extern void init_back();
extern int init_front(char *errors, int pointsallowed, int polysallowed );


/* *******************  functions  ******************************* */

void sync();



void whatplane( char * choice ){  /* ask user what plane he/she wants to fly */

struct ffblk files;
char temps[40][15];
int i,k;
char *p;


   if( findfirst("*.par",&files,FA_NORMAL) ) return;  /* none found */
   i= 1;
   strcpy( temps[0], files.ff_name );
   p= temps[0];
   while ( *p != '.' ) ++p;
   *p= 0;                   /* chop off the file extension */

   while ( findnext(&files ) == 0 ){
      strcpy( temps[i],files.ff_name );
      p= temps[i];
      while (*p != '.' ) ++p;
      *p= 0;
      ++i;
      }

  /* we now have all the names, so ask the user which one */
#ifdef  TURBO
  clrscr();
#else
  clrscrn();
#endif
  printf("Welcome to FlyRc.\nWritten by Ron Carr\n\n\n");
  for( k= 0; k < i; ++k ){
     printf("  %d   -   %s\n", k+1, temps[k]);
     }
  while ( k < 0 || k >= i ){
    printf("\n  Enter a number to select a plane to fly..... ");
    k= getch() - '1';
    }

  strcpy ( choice, temps[k] );

}

/*
 adjust the background pixel position based upon the x value of position
 then rotate the coordinate system so x is zero again
 value of pix is adjusted to its new value making the background move
*/
void findpix(){
double angle;
int inc;
double adder;
double x,y,z;
double dist;
double temp;

  x= position.x - view_position.x;
  y= position.y - view_position.y;
  z= position.z - view_position.z;

  /* find the new xyz for the view co-ordinates */
  dist= sqrt( x*x + y*y );
  /*
    look directly at our plane - this makes the y axis the same as the
    position vector, so just normalize the position.
  */
  view[0][1]= -x/dist;   /* but this works backward */
  view[1][1]= y/dist;
  /* the x axis is 90 deg from the y axis */
  view[0][0]= view[1][1];
  view[1][0]= -view[0][1];

  /* need this if appending any rotations afterwards */
  view[0][2]= view[1][2]= view[2][0]= view[2][1]= 0.0;
  view[2][2]= 1.0;

  temp= (int)( 408.0 * acos( view[0][0] ));
  if( x < 0 ) pix= 800 - (int)temp;
  else pix= 800 + (int)temp;

  if( pix < 0 ) pix+= 2560;
  if( pix > 2559 ) pix-= 2560;

return;



}

/* determine how much we tilt our heads */
void findbackline(){
double dist;
double x,y,z;

/* this is redundant code */
  x= position.x - view_position.x;
  y= position.y - view_position.y;
  z= position.z - view_position.z;



   dist= sqrt( x*x + y*y + z*z );
/* view angle will be from 90 deg to -90 deg */
   viewangle= asin( z/dist );
   backline= (int)( viewangle * 400/1.57 ) ;

   if( backline > 400 ) backline= 400;
   else if( backline < -400 ) backline= -400;

/* tilt our view by some fraction of the actual allowing plane to float up/down */
   AppendRotationX(view, -0.6 * viewangle );


 /* calculate the sliding window for offscreen image */
/*  on ground - close by - at normal distance */

   if( z < -8 ) row_offset= 339;
   else if( y < 80 ) row_offset= 410 + ( int )( 0.35 * (double)backline );
   else row_offset= 339 + (int)(0.5  *(double)backline );
   if( row_offset > 480 ) row_offset= 480;
   if( row_offset < 339 ) row_offset= 339;

}


void moveplane(){

/* pull velocity vector out of the xform */
/* move in direction of the plane's y axis */

    position.x+= xform[0][1] * speed;
    position.y+= xform[1][1] * speed;
    position.z+= xform[2][1] * speed;

}



void flyplane(){
double xroll, yroll, zroll;
double xchange;
double signed_drag;
double myspeed;
double diffspeed;



/* see if speed should change */
/* resultant_thrust is power - drag +- gravity  */
/* sign to fix speed increase due to negative drag when going backward */

   myspeed= ( speed - best_speed );
   myspeed= myspeed * myspeed;
   if( speed > best_speed ) signed_drag= drag_floor + drag * myspeed;
   else signed_drag= drag_floor + lowdrag * myspeed;
   if( position.z < -9.9 ) signed_drag*= 2.0;    /* on ground */
   if ( speed < 0 ) signed_drag= -signed_drag;

/* and speed changes by result_thrust * inertia ( thrust variable is power )*/
   speed= speed +  (thrust - signed_drag - xform[2][1]*gthrust)*inertia;

/* figure torque's */

/* yroll is aileron */
/* note key effect was hard coded -  was 0.03  now called keyrate */
   yroll= (double)ajoy * aspeed * speed + (double)atrim * keyrate;

/* gtrim simulates wing lift, if wings level then xchange will equal gtrim */
/* ztrim turns the plane toward the ground when wings not level */

   if( (diffspeed= stall - speed ) > 0 && position.z > -9.9 ){       /* speed < stall */
      position.z-= sdrop * diffspeed;
      xchange=  xform[2][2] * ( gtrim + gstall * diffspeed );
      zroll= xform[2][0] * ( ztrim + zstall * diffspeed );
      }
   else{    /* normal model */
      xchange= xform[2][2]*gtrim;
      zroll= xform[2][0]*ztrim;
      }
   if( speed > 0.1 ) zroll+= (double)rudder * rspeed * thrust * 20 / speed;

   xroll= (double)ejoy * espeed * speed + gtrim - xchange
        +(double)etrim * keyrate;

/* apply torque */
   appendz(xform,zroll);
   appendx(xform,xroll);
   appendy(xform,yroll);

   etrim= atrim= 0;

   moveplane();

}




/* project and draw the shadow's, arg i is index to poly array */
/* now have true projection of the shadow points */
void draw_shadow(int i){
int j;
struct POLY temp;
struct POINT tpoint[POINTSINPOLY];
int tvert[POINTSINPOLY];
int *p;

   temp.color= SHADOW;
   temp.verts= tvert;

   /* copy the poly to temp storage */
   temp.no_points= polys[i].no_points;
   p = polys[i].verts;
   for( j= 0; j < temp.no_points; ++j, ++p ){
   /* copy and project the points */
       tvert[j]= j;
       tpoint[j].x= points[*p].x;
       tpoint[j].y= points[*p].y;
       tpoint[j].z= points[*p].z;   /* shadow always on the ground */
       project( &tpoint[j], 1);
       }

   drawpoly(&temp,&tpoint);   /* draw the poly */

}


/* draw the polys back to front, skip shadow polys */
void drawinorder(){
int i,j,k;
double order[MAX_POLY];
int * vert;
double sum;


   planemask(3);

 /* determine the ave distance of each poly */
   for( i= 0; i < no_polys; ++i ){
      if( polys[i].color == SHADOW ){
         draw_shadow(i);
         order[i]= 0.0;    /* get float NaN without initializing doubles */
         continue;
         }
      vert= polys[i].verts;
      k= j= polys[i].no_points;
      sum= 0.0;
      while( j-- ) sum+= points[*vert++].dist;
      order[i]= sum/(double)k;
    /* clip close objects, .0025 is projection factor -> .0075 for 3 feet */
      if( order[i] < 0.01 ) order[i]= 0.0;
      }

 /* draw the farthest first */
   k= 0;
   while( k >= 0 ){
      k= -1;
      sum= 0.0;
      for( i= 0; i < no_polys; ++i ){     /* dumb way */
         if( order[i] > sum ){
            sum= order[i];
            k= i;
            }
         }
      if( k >= 0 ){
         if( polys[k].color != SHADOW ) drawpoly(&polys[k],&points);
         order[k]= 0.0;
         }
      }

}

void project( struct POINT *p , int shadow_flag ){
struct POINT dest,dest2;
double dist;

/*    if( runway_flag == 0 ){ */

    /* rotate the object into world space */
    xformpoint(xform,p,&dest);
    /* translate point to world space */
    dest.x+= position.x;
    dest.y+= position.y;
    dest.z+= position.z;
    if( shadow_flag ) dest.z= -10;

/*       }
    else{
       dest.x= p->x;  dest.y= p->y;  dest.z= p->z;
       }    */

    /* translate to view origin */
    dest.x+= -view_position.x;
    dest.y+= -view_position.y;
    dest.z+= -view_position.z;   /* not really needed yet */

    /* project point to view space */
    xformpoint(view,&dest,&dest2);

    p->dist= sqrt (dest2.x*dest2.x + dest2.y*dest2.y + dest2.z*dest2.z);
/*    p->ry= dest2.y;  */
    /* fixed projection factor 0025 based upon number of pixels in background*/
    dist= 0.0025 * p->dist;

    /* determine screen co-ordinates */
    p->col= (int) (dest2.x/dist) + 320;
    p->row= 240 - (int)(dest2.z/dist);

}




/* start in a know position */
void init_flight(){
int i;

   etrim= atrim= 0;
   speed= thrust= 0.0;
   for( i= 0; i < 3; ++i ){
      xform[i][0]= xform[i][1]= xform[i][2]= 0;
      }
   xform[0][0]= 1;
   xform[1][1]= 1;
   xform[2][2]= 1;
   appendz(xform,-0.157);

   position.x= 0;
   position.y= 15;
   position.z= -10;

 /*  pix= 800; */

   findpix();
   findbackline();
   draw_back(879+backline,pix);
   draw_front();

}

input_setup(){  /* set up input interface */

   switch( iiface ){
      case 0:    /* keys */
         espeed= aspeed= rspeed= 0.0;
         break;
      case 1:    /* joystick */
         rspeed= espeed= ejoyspeed;
         aspeed= ajoyspeed;
         rspeed/= 1.0;
         break;
      case 2:    /* radio */
         rspeed= espeed= eradiospeed;
         aspeed= aradiospeed;
         rspeed/= 1.0;
         break;
      }
}


int keys(){    /* handle key input, return non zero for program exit */
int c;


   while( kbhit() ){
      c= tolower(getch());
      switch( c ){
         case 'j':  iiface= 1 - nojoy;  input_setup(); break;
         case 'k':   iiface= 0;  input_setup(); break;
         case 'r':  iiface= rcsticks;  input_setup(); break;

         case 'q':  case 27:   return 1;    /* program exit */
         case 'i':  init_flight();  break;

         case 'f':                /* adjust stick rate */
            aspeed+= aspeed/10.0;
            espeed+= espeed/10.0;
            break;
         case 's':
            aspeed-= aspeed/10.0;
            espeed-= espeed/10.0;
            break;

         case '1' : case '2':  case '3':  case '4':
         case '5' : case '6':  case '7':  case '8': case '9':
            thrust= (double)( c - '1' )*power;
            break;
         case 0:
         c= getch();
            switch ( c ){
            case DOWN:   /* down arrow is up elevator */
               if( etrim < 64 ) ++etrim;
               break;
            case UP:
               if( etrim > -64 ) --etrim;
               break;
            case LEFT:
               if( atrim > -64 ) --atrim;
               break;
            case RIGHT:
               if( atrim < 64 ) ++atrim;
               break;
            case 71:   /* home, move toward plane */
               view_position.y+= 2*view[1][1];
               view_position.x-= 2*view[0][1];
               break;
            case 79:   /* end, move away from plane */
               view_position.y-= 2*view[1][1];
               view_position.x+= 2*view[0][1];
               break;
            case 83:   /* delete, move sideways to left of plane */
            /* to stay at original distance, need to move towards
               plane a small amount after going sideways ( now the
               viewer moves on a helix away from the plane ) */
               view_position.y-= 2*view[0][1];
               view_position.x-= 2*view[1][1];
               break;
            case 81:   /* page down, move sideways to right of plane */
               view_position.y+= 2*view[0][1];
               view_position.x+= 2*view[1][1];
               break;
            break;
            }  /* end switch special key */

         }  /* end key switch */
      }

   return 0;
}



void draw_front(){      /* draw the plane */
int i;

      for( i=0; i < no_points; ++i ) project(&points[i], 0);
      drawinorder();
      sync();
      dumpscan( );

}

main(int argc, char *argv[]){
int x;
double y;
int i,c;
char temp[40],err[256],line[256];
char temp2[40];
int frame_count;
time_t ticks;
FILE *params;
int exitflag;
int dflag= 0;       /* if we want message at the end */
int plane_arg= 0;   /* if found plane name */

/* ask what plane to fly if no arguments */

   for( i= 1; i < argc; ++i ){
      if( argv[i][0] == '-' ){     /* it's a switch */
         switch( argv[i][1] ){
           case 'd':  dflag= 1; break;
           case 't':  /* tx transmitter values */
             ++i;  x= atoi( argv[i] );
             status= x+1;  control= x+2;
             ++i;  invert= atoi(argv[i] );
             ++i;  nominal= atoi(argv[i] );
             rcsticks= 2;
             outportb(control,0x4);  /* energize pullup resistor */
           break;
           }
         }
      else {         /* we have a plane name */
         strcpy(temp,argv[i]);
         strcpy(temp2,argv[i]);
         plane_arg= 1;
         }
      }

   if( plane_arg == 0){
      if( dflag == 1 ){     /* no plane selected */
         whatplane( temp );
         strcpy( temp2, temp);
         }
      else execl("flyrc",0);   /* force new intro screen if not debugging */
      }

   strcat(temp,".dat");
   strcat(temp2,".par");

   vmode(0x12);  /* 640x480 16 colors */

/* get the assignment of channels to aileron elevator throttle rudder */
params= fopen("radio.cfg","r");
if( params ) {
    fscanf(params,"%d%d%d%d",&av,&ev,&tv,&rv);
    }
else{
   av= 1; ev= 2; tv= 3; rv= 4;
   }
fclose(params);


/* read in the flight contants that we vary ;-> */
   params= fopen(temp2,"r");
   if( params ){
         fscanf(params,"%lf",&power);
         fgets(line,256,params);
         fscanf(params,"%lf",&drag);
         fgets(line,256,params);
         fscanf(params,"%lf",&lowdrag);
         fgets(line,256,params);
         fscanf(params,"%lf",&inertia);
         fgets(line,256,params);
         fscanf(params,"%lf",&gthrust);
         fgets(line,256,params);
         fscanf(params,"%lf",&gtrim);
         fgets(line,256,params);
         fscanf(params,"%lf",&ztrim);
         fgets(line,256,params);
         fscanf(params,"%lf",&sdrop);
         fgets(line,256,params);
         fscanf(params,"%lf",&stall);
         fgets(line,256,params);
         fscanf(params,"%lf",&ejoyspeed);
         fgets(line,256,params);
         fscanf(params,"%lf",&ajoyspeed);
         fgets(line,256,params);
         fscanf(params,"%lf",&keyrate);
         fgets(line,256,params);
         fscanf(params,"%lf",&best_speed);
         fgets(line,256,params);
         fscanf(params,"%lf",&drag_floor);
         fgets(line,256,params);
         fscanf(params,"%lf",&gstall);
         fgets(line,256,params);
         fscanf(params,"%lf",&zstall);
         fgets(line,256,params);
         fscanf(params," %x %x",&p1c,&p2c);    /* hate scanf sometimes */
         fclose(params);
         }


  rspeed= espeed= eradiospeed= ejoyspeed;
  aspeed= aradiospeed= ajoyspeed;
  rspeed/= 1.0;

/* set up plane colors */
   mypal[2]= mypal[6]= mypal[10]= mypal[14]= p1c;
   mypal[3]= mypal[7]= mypal[11]= mypal[15]= p2c;

   setmypal(mypal);

   writemode(3);    /* uses write mode3 to set pixels */
   planemask(0xf);

   err[0]= 0;       /* need to init if not used */
   init_back();
   if( init_front(temp,MAX_POINT,MAX_POLY)){
       sprintf(err,"Error in file %s\n",temp);
       goto end;
       }

   temp[0]= 0;




/* do a 3 sec fly by to determin the processor frame rate */

   frame_count= 0;
   position.x= -100.0;       /* starting position */
   position.y= 80.0;
   position.z=  10.0;

   for(i= 0; i < 8; ++i )appendz(xform,-0.2);   /*.281 */
   ticks= clock();

/* calibrate joystick */

   if( (inportb(0x201) & 3 ) == 0 ){
      iiface= 1;
      nojoy= 0;
      sync();
      ajoycal= joy.ax;
      ejoycal= joy.ay;
   /* adjust speed of joystick to calibrated value */
      ejoyspeed= 64.0 * ejoyspeed /(double)ejoycal;
      ajoyspeed= 64.0 * ajoyspeed /(double)ajoycal;
      input_setup();
      }

   while( clock()-ticks < CLK_TCK * 3){
      ++frame_count;
      moveplane();
      findpix();
      findbackline();
      draw_back(879+backline,pix);
      draw_front();
      }


 /*   break plane and crash */
x= 0; y= 0.0;

while( kbhit() == 0 ){
   if( x == 0 && position.z < y ){
       x= 1;
       }
   if( x == 1 ){
     appendy(xform,-0.031);
     position.z-= 0.2;
     }
   y= position.z;

   if( position.z < -10 ) break;
   appendx(xform,0.03);
   moveplane();
   findpix();
   findbackline();
   draw_back(879+backline,pix);
   draw_front();
   }

/* set up for normal flight - previous was just demo */
/* we now know our frame rate, frames per sec (after 3 seconds) */
   frames= (double)frame_count/3.0;

 if( frames < 30 ) {   /* help the 286 class machine */

   power/= 60;
   drag/= 60;
   lowdrag/= 60;
   inertia/= frames;
   gtrim/= 60;
   ztrim/= 60;
   gthrust/= 60;
   sdrop/=  60;
   stall/= 60;
   }


else {
/* assume 60 frames per second and adjust speed in a different manner */
   power/= 60.0;
   drag/= 60.0;
   lowdrag/= 60.0;
   inertia/= 60.0;
   gtrim/= 60.0;
   ztrim/= 60.0;
   gthrust/= 60.0;
   sdrop/=  60.0;
   stall/= 60.0;
   }

/* new parameters */
  best_speed/= 60.0;
  drag_floor/= 60.0;
  gstall/= 60.0;
  zstall/= 60.0;


getch();

   init_flight();
   exitflag= 0;

   if( rcsticks ){
       eradiospeed= 3.0 * 64.0 * eradiospeed / (double)nominal;
       aradiospeed= 3.0 * 64.0 * aradiospeed / (double)nominal;
       iiface= 2;  /* assume want to use rcbox as input */
       input_setup();
       }

   /*  main  program loop */
   while(exitflag == 0){
      if( keys() ) break;   /* escape or quit */
      flyplane();

    /* detect crash */
      if( position.z < -10 ){
         if( speed > ( stall + 0.4) || xform[2][2] < 0.9 ){
        /* speed is 1/60th of what you would expect */
        /* hold crash in view and flush keys */

            ticks= clock();
            while( clock()-ticks < (time_t)(CLK_TCK * 5)){ /* just delay 5 seconds */
               if( kbhit()){
                  c= tolower(getch());
                  if( c == 'i' || c == 27) break;
                  }
               }
            init_flight();
            }
         else{      /* landed so just keep going */
            position.z= -10.0;
            appendx( xform, -asin( xform[2][1] ) );
       /* can't do this this way now,  xform[2][2]= 1.0;  /* level plane */
       /*      appendz(xform,0);   */

            }
         }

      findpix();
      findbackline();
      draw_back(879+backline,pix);
      draw_front();

      }  /* end main loop */



end:
   vmode(3);   /* return to text mode */
   printf("%d pictures in 3 sec.  Frame rate is  %4.2lf/sec ",frame_count,frames);
   printf("\nspeed is %lf (if rate >= 60 frames/sec)",speed*60.0);
   printf("\nJoystick %s found",nojoy ? "not" : "" );
   printf("\nRadio interface %s found",rcsticks ? "" : "not");

   if( dflag ){
      printf("\nviewangle is %lf",viewangle);
      printf("\nPosition is X %lf    Y %lf     Z %lf\n",position.x,position.y,position.z);
      printf("Plane XFORM\n");
      for( i= 0; i < 3; ++i ){
         printf("%4.3lf %4.3lf %4.3lf\n",xform[i][0],xform[i][1],xform[i][2]);
         }

      printf("View XFORM\n");
      for( i= 0; i < 3; ++i ){
         printf("%4.3lf %4.3lf %4.3lf\n",view[i][0],view[i][1],view[i][2]);
         }

      }
   printf("\n%s",err);   /* print any errors */
}

int count_high(){
int count;
int val;
   do{
      val= (inportb(status) & 0x40 ) ^ invert;
      }while( val == 0 );
   count= 0;
   while( ( inportb(status) & 0x40 ) ^ invert ){
       ++count;
       }
   return count;
}

void sync(){    /* waits for vertical sync and reads joystick */
int c, saw_sync,x;
int count;
int count2;
double val,norm;
int counts[5];

    joy.ax= joy.ay= saw_sync= count= 0;

    if( iiface == 1 ){      /* using joystick */
       outportb(0x201,0xff);
       while( ++count < 10000 ){
          c= inportb(0x201);
          if( c & 1 ) ++joy.ax;
          if( c & 2 ) ++joy.ay;
          if( inportb(INPUT_STATUS1 ) & VSYNC_MASK ) saw_sync= 1;
          if( (c & 3) == 0 ) break;
          }
       ejoy= ( joy.ay - ejoycal );
       ajoy= ( joy.ax - ajoycal );
       }
    else if( iiface == 2 ){    /* using radio */
       saw_sync= 1;   /* this is probably slow */
                      /* sync to radio and not to vertical refresh */
       for( x= 0; x < 10; ++x ){   /* find sync */
          count2= count_high();
          if( count2 > 2 * nominal ) break;
          }
       if( x != 10 ){
          for( x= 1; x < 5; ++x ) counts[x]= count_high();
          ajoy= ( counts[av] - nominal );
          ejoy= -( counts[ev] - nominal );
          val= (double)counts[tv];   norm= (double) nominal;
          thrust= ( val/norm - 0.6666 ) * power * 15.0;
       /* add rudder as counts[rv] */
          rudder= -( counts[rv] - nominal );
          }
       }

                           /* return;   /* raw speed indicator */
    if( saw_sync == 0 )
           while( (inportb(INPUT_STATUS1) & VSYNC_MASK) != VSYNC_MASK );
    return;
}


