#include "poly.h"
#include "graph.h"
#include "pallette.h"
#include <stdio.h>
#include <malloc.h>
#include "vga.h"
#include <dos.h>

extern int no_points;
extern int no_polys;
extern struct POINT points[];
extern struct POLY polys[];


/*
   screen memory is divided into 480 visible rows and
   339 rows that are off screen.  The forground objects ( plane and shadow )
   are first drawn in off screen memory implimenting the painters algorithm.
   Then they are copied to the visible screen.  If more than 339 lines are
   needed, the forground objects are clipped to 339 lines.
*/

/* value should be between 340 and 480 to cover all of the screen */
extern int row_offset;   /* value is function of backline */

/* storage for forground horizontal lines */
/* index to the array is the scanline */
struct ENTRY{
   int start;   /* start pixel */
   int end;     /* end pixel */
   int erase_start;
   int erase_end;
   int draw_start;
   int draw_end;
   int done;     /* for direct to screen shadows */
   };

#define NOFRONTLINES 480
static struct ENTRY fscanline[NOFRONTLINES];

static int low_row;   /* limit of current poly */
static int high_row;  /* limit of current poly */


/*
  add the point to our scanline array - the color is gcolor
  changed this routine to extend existing lines for current poly
  solving the single point in the array problem
*/
void addpoint( int col, int row ){

   /* clip to screen */
   if( row < 0 || row > 479 ) return;
   if( col < 0 ) col = 0;
   if( col > 639 ) col= 639;

   /* save limits of this poly */
   if( row < low_row ) low_row= row;
   if( row > high_row ) high_row= row;

   /* check if this is the first point, it so then store as both start and end */
   if( fscanline[row].start == 640 ){
      fscanline[row].start= fscanline[row].end= col;
      return;
      }

   /* save this value */
   if( col < fscanline[row].start ) fscanline[row].start= col;
   else if( col > fscanline[row].end ) fscanline[row].end= col;

}


/*
   Draw a line in specified gcolor using Bresenham's integer based algorithm.
   except in this case we enter the points into our scanline array

   modified traceline to use a function pointer to the function that adds
   the points to the data structure - now can use this function to scan
   the runway lines

   changed the algorithm slightly - never add first point of line, always
   add last point, single points no longer a problem because we are checking
   that the points belong to the correct poly
 */
void traceline( struct POINT *p1, struct POINT *p2, void (*addpoint)() ){
  register int t, distance;
  int x=0, y=0, delta_x, delta_y;
  int incx, incy;
  int startx, starty, endx, endy;

  startx= p1->col;  starty= p1->row;
  endx= p2->col;  endy= p2->row;


  /* Compute the distances in both directions. */
  delta_x = endx-startx;
  delta_y = endy-starty;

  if( delta_y == 0 ){    /* horizontal line is special case */
     (*addpoint)(endx,endy);
     return;
     }

  /* Compute the direction of the increment.
     An increment of 0 means either a vertical
     or horizontal line.
  */
  if(delta_x>0) incx = 1;
  else if(delta_x==0) incx = 0;
  else incx = -1;

  if(delta_y>0) incy = 1;
/*  else if(delta_y==0) incy = 0;   already tested for this condition */
  else incy = -1;

  /* determine which distance is greater */
  delta_x = (delta_x >= 0) ? delta_x : -delta_x;
  delta_y = (delta_y >= 0) ? delta_y : -delta_y;
  if(delta_x>delta_y) distance = delta_x;
  else distance = delta_y;

  /* draw the line */

  for(t=0; t<=distance; t++) {
    x += delta_x;
    y += delta_y;
    if(x>distance) {
      x -= distance;
      startx += incx;
    }
    if(y>distance) {
      y -= distance;
      starty += incy;
      (*addpoint)(startx, starty);
      }
    }

}


void drawpoly( struct POLY *poly, struct POINT *point ){
int i,j;
int *v;
struct POINT *p0, *p1, *p2;
int row;

   low_row= 480;   /* reverse setup these limits */
   high_row= 0;

/*   gcolor= poly->color; */
   j = poly->no_points;
   v= poly->verts;
   p0= p1= &point[*v++];

   for( i= 1; i < j; ++i ){
      p2= &point[*v++];
      traceline( p1, p2,&addpoint );
      p1= p2;
      }

   traceline(p2,p0,&addpoint);     /* close loop with lastpoint to first */

/* now need to dump the poly to non-visible screen memory */
/* this call to planemask should be outside of this loop, maybe in main */
   planemask(3);    /* write to foreground */

   mypencolor( poly->color );

   for( j= low_row; j <= high_row; ++j ){
    /* j is visible row - j+row_offset is offscreen */
      row= j+row_offset;
      if( row > 479 && row < 819 ){  /* clip to offscreen video memory */
         hline3( fscanline[j].start , row , fscanline[j].end - fscanline[j].start + 1 );
         if( fscanline[j].start < fscanline[j].draw_start ) fscanline[j].draw_start= fscanline[j].start;
         if( fscanline[j].end > fscanline[j].draw_end ) fscanline[j].draw_end= fscanline[j].end;
         }
      else if( poly->color == 1 ) {   /* draw just shadow direct to screen */
         /* first erase what is there */
         if( fscanline[j].done == 0 ){
            if( fscanline[j].erase_start != 640 ){
               mypencolor( 0 );
               hline3( fscanline[j].erase_start, j, fscanline [j].erase_end - fscanline[j].erase_start  + 1);
               mypencolor( 1 );
               }
            fscanline[j].done= 1;
            }
         /* then draw this segment */
         hline3( fscanline[j].start, j, fscanline[j].end - fscanline[j].start + 1);
         if( fscanline[j].start < fscanline[j].draw_start ) fscanline[j].draw_start= fscanline[j].start;
         if( fscanline[j].end > fscanline[j].draw_end ) fscanline[j].draw_end= fscanline[j].end;
         }
      fscanline[j].start=640;  fscanline[j].end= 0; /* reset these */
      }
}

/* dumpscan now needs to copy the offscreen image into the visible area */
void dumpscan(){
int i,j;
int start,end;


   mypencolor( 0 );
   i= row_offset;
   for( j= 0;  j < 480; ++j, ++i ){
      /* j= i-row_offset;   */
      /* i= j+row_offset;   */
      start= fscanline[j].erase_start;
      if( fscanline[j].draw_start < start ) start= fscanline[j].draw_start;
      if( start == 640 ) continue;    /* no change needed on screen */
      end= fscanline[j].erase_end;
      if( fscanline[j].draw_end > end ) end= fscanline[j].draw_end;
      fscanline[j].erase_start= fscanline[j].draw_start;
      fscanline[j].erase_end= fscanline[j].draw_end;
      fscanline[j].draw_start= 640;
      fscanline[j].draw_end= 0;
    /* if image is offscreen then copy */
      if( i > 479 && i < 819 )  copyline(start, j, end-start+1, i );
      else if( fscanline[j].done ) fscanline[j].done= 0;   /* did this line direct to screen */
    /* out of bounds, then just erase on screen directly */
      else hline3(start,j,end-start+1);
      }



/* debug - show limits of possible external screen area */
/*
mypencolor(2);
hline3(100,480-row_offset,440);
mypencolor(3);
hline3(100,819-row_offset,440);
sync();
mypencolor(0);
hline3(100,480-row_offset,440);
hline3(100,819-row_offset,440);
*/

}


int init_front(char *file, int max_points, int max_polys){
FILE *fp;
char temp[256];
int j;
int color;
int noints;
int *v;
double scale;

    /* read in the plane datafile */
   fp= fopen(file,"r");
   if(fp == 0 ) return 1;

   /* read points first */
   if( fscanf(fp,"%d",&no_points) == EOF ) return 1;
   fgets(temp,256,fp);   /* strip comments */
   if( fscanf(fp,"%lf",&scale) == EOF ) return 1;
   fgets(temp,256,fp);   /* strip comments */
   if( no_points > max_points ) return 1;
   for( j= 0; j < no_points; ++j ){
      fscanf(fp,"%lf%lf%lf",&points[j].x,&points[j].y,&points[j].z);
      points[j].x*= scale;
      points[j].y*= scale;
      points[j].z*= scale;
      }

   /* read poly's */
   if( fscanf(fp,"%d",&no_polys) == EOF ) return 1;
   fgets(temp,256,fp);   /* strip comments */
   if( no_polys > max_polys ) return 1;

   for( j= 0; j < no_polys; ++j ){
      fscanf(fp,"%d",&color);
      fgets(temp,256,fp);
      fscanf(fp,"%d",&noints);
      fgets(temp,256,fp);
      polys[j].color= color;
      polys[j].no_points= noints;
      v= (int *) malloc( sizeof(int) * noints );
      if( v == 0 ) return 1;
      polys[j].verts= v;
      while(noints--){
         fscanf(fp,"%d",&color);   /* reuse color var for temp storage */
         *v++= color;
         }
      }

fclose(fp);

/* setup the scanline array with reversed limits */
   for( j= 0; j < NOFRONTLINES; ++j ){
      fscanline[j].draw_start= fscanline[j].erase_start= fscanline[j].start= 640;
      fscanline[j].erase_end= fscanline[j].end= fscanline[j].draw_end= 0;
      }

return 0;   /* no error */
}


