

#include <stdlib.h>
#include <graphics.h>
#include <bios.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include "vga.h"
#include <malloc.h>


void vmode(int mode_code);
void writepix2(int x, int y );   /* mode 2 */
void writepix3(int x, int y );   /* mode 3 */
void setmypal(int *pal);        /* array of 16 values */
void mylineto(int x, int y );
void mymoveto( int x, int y );
void planemask( int planes );
void writemode( int x );
void mypencolor(int color);
void enable_setreset(int x);
void setreset(int x);
void hline3(int x, int y, int bits);

static int graphx, graphy;
static int pencolor;
static int using_mode;

void mymoveto( int x, int y ){

   graphx= x;    graphy= y;
}

void mypencolor( int color ){

   if( color == pencolor) return;
   pencolor= color;
   if( using_mode == 3 ) setreset(color);     /* color for mode 3 */

}

void setreset( int color ){

      outportb( GC_INDEX,GC_SET_RESET );
      outportb( GC_DATA,color );   /* set/reset has color */

}


void enable_setreset( int planes ){

   outportb( GC_INDEX,GC_ENABLE_SET_RESET );
   outportb( GC_DATA,planes );


}

/*   set ega pallette */
void setmypal(int *pal){

union REGS regs;
int i;


   for( i= 0; i < 16; ++i ){   /* just set each in turn to avoid segment stuff */
      regs.h.ah= 0x10;    /* vid bios call */
      regs.h.al= 0;       /* set one pallette reg */
      regs.h.bl= (char)i;       /* reg to set */
      regs.h.bh= (char)pal[i];  /* value to set */
      int86(0x10,&regs,&regs);
      }
}



/* Set the video mode. */
void vmode(int mode_code)
{
  union REGS r;

  r.h.al = (char)mode_code;
  r.h.ah = 0;
  int86(0x10, &r, &r);


}


/*
   Draw a line in specified color using Bresenham's integer based algorithm.
   requires write mode 3 be set previously
*/
void mylineto( int endx, int endy ){
  register int t, distance;
  int x=0, y=0, delta_x, delta_y;
  int incx, incy;
  int startx, starty;

  startx=graphx;   starty=graphy;

  /* Compute the distances in both directions. */
  delta_x = endx-startx;
  delta_y = endy-starty;

  /* Compute the direction of the increment.
     An increment of 0 means either a vertical
     or horizontal line.
  */
  if(delta_x>0) incx = 1;
  else if(delta_x==0) incx = 0;
  else incx = -1;

  if(delta_y>0) incy = 1;
  else if(delta_y==0) incy = 0;
  else incy = -1;

  /* determine which distance is greater */
  delta_x = abs(delta_x);
  delta_y = abs(delta_y);
  if(delta_x>delta_y) distance = delta_x;
  else distance = delta_y;

  /* draw the line */
  for(t=0; t<=distance+1; t++) {
    writepix3(startx, starty);
    x += delta_x;
    y += delta_y;
    if(x>distance) {
      x -= distance;
      startx += incx;
    }
    if(y>distance) {
      y -= distance;
      starty += incy;
    }
  }


graphx= endx;   graphy= endy;   /* set new graphics cursor position */

}


/* Write a pixel directly to graphics RAM using write mode 2 */
/* in interest of some speed - requires mode 2 be set before calling */
/* this is a 640 column mode - 80 bytes across */
void writepix2(int x, int y){

char far * video;
unsigned offset;
int bitmask;
int i;
char c;

   offset= 80*y+(x >> 3);        /* determine bit position */
   i= x & 7;
   bitmask= 0x80 >> i;
   video= MK_FP( 0xa000, offset );

/* set the bit mask */
   outportb( GC_INDEX,BIT_MASK);
   outportb( GC_DATA,bitmask );

/* set the latches and write the bit */
   c= *video;
   *video= pencolor;

}

/*
    write pixel using writemode 3
    the set/reset( color ), bitmask(0xff) and mode 3 should be set before
    calling this function
    for  640x350 mode - or other 640x?? modes
*/
void writepix3(int x, int y){

char far * video;
unsigned offset;
int bitmask;
int i;
char c;

   offset= 80*y+(x >> 3);        /* determine bit position */
   i= x & 7;
   bitmask= 0x80 >> i;
   video= MK_FP( 0xa000, offset );

/* set the latches and write the bit */
   c= *video;
   *video= bitmask;

}


/* horizontal line in write mode 3 */
void hline3(int x, int y, int bits){

char far * video;
unsigned offset;
int mask;
register int i;
char c;

static int fixmask[8]={       /* precalculated masks for final write */
  0,0x80,0xc0,0xe0,0xf0,0xf8,0xfc,0xfe };

/*   bits+= 1;               /* write end point also */
   offset= 80*y+(x >> 3);        /* determine 1st bit position */
   i= x & 7;
   video= MK_FP( 0xa000, offset );

   mask= 0xff >> i;
   if( bits + i <= 8 ){                /* just need one write ?? */
      mask= mask ^ ( mask >> bits );
      c= *video;
      *video= mask;
      return;
      }

   bits= bits + i - 8;          /* else more bits in next address */

/* set the latches and write the 1st partial bits */
   c= *video;
   *video++= mask;

/* write 8 bits at a time, read into latches not needed */
   i= bits >> 3;

   while( i-- ) *video++= 0xff;

/* now write any leftover bits */
/* use fixed mask */
   c= *video;
   *video= fixmask[bits & 7];

}

/* copy horizontal line in write mode 3 */
/* we can use set/reset to clear unwanted bits because the pallette is
   stacked to separate the bit planes - only 2 planes are enabled */
/* copyline had a problem when the limits of the line spanned some
   leftover color in the middle, changed erase algorithm */
/* changed to also erase the source - in blocks of 8 bits ( erase not masked) */
void copyline(int x, int y, int bits, int source_row){

char far * video;
char far *offscreen;
unsigned offset,offset2;
int mask;
register int i;
char c;

static int fixmask[8]={       /* inverted precalculated masks for final write */
  0xff,0x7f,0x3f,0x1f,0x0f,0x07,0x03,0x01 };

/*   mypencolor(0);          /* use set/reset to clear on the edges */
/*   bits+= 1;               /* write end point also */

   offset= 80*y+(i= (x >> 3));        /* determine 1st bit position */
   offset2= 80*source_row +i;                /* source position, just offscreen */
   i= x & 7;
   video= MK_FP( 0xa000, offset );
   offscreen= MK_FP( 0xa000, offset2 );
   mask= 0xff >> i;
/* mask is inverted because we want to use latched data */
   if( bits + i <= 8 ){                /* just need one write ?? */
      mask= mask ^ ( mask >> bits );
      c= *offscreen;                   /* copy the line via latches */
      *video= (mask ^ 0xff);
      *offscreen= 0xff;
      return;
      }

   bits= bits + i - 8;          /* else more bits in next address */

/* set the latches and write the 1st partial bits */
   c= *offscreen;
   *offscreen++= 0xff;
   *video++= mask ^ 0xff;

/* write 8 bits at a time */
   i= bits >> 3;
   while( i-- ){
      c= *offscreen;
      *offscreen++= 0xff;
      *video++= 0x00;       /* zero bit mask uses just the latched data */
      }

/* now write any leftover bits */
/* use fixed mask */
   c= *offscreen;
   *offscreen= 0xff;
   *video= fixmask[bits & 7];

}



void planemask( int planes ){

/* set the plane mask for writing */
   outportb( SC_INDEX, SC_MAP_MASK );   /* index mask */
   outportb( SC_DATA, planes );

}


void writemode( int mode ){    /* set the graphics write mode */
unsigned char c;

   outportb( GC_INDEX,GC_MODE );
   c= inportb( GC_DATA ) & 0xfc;
   outportb( GC_DATA, c | mode );
   outportb( GC_INDEX,BIT_MASK);    /* enable all bits for mode 3 */
   outportb( GC_DATA,0xff );        /* enable all bits any mode */
   using_mode= mode;


}


