#include <math.h>


/*
   Routines to perform incremental rotations around the three axes

   Changed from original to increment about the new object coordinates vs
   the original world coordinates.  This allows the plane to roll about
   its own axis vs some predefined fixed axis
*/


/* Matrix multiplies Xform by SourceVec, and stores the result in
   DestVec. Multiplies a 3x3 matrix times a 3x1 matrix; the result
   is a 3x1 matrix, as follows:
   --     --     -- --     -- --
   |       |     | 3 |     | 3 |
   |  3x3  |  X  | x |  =  | x |
   |       |     | 1 |     | 1 |
   --     --     -- --     -- -- */
void xformpoint(double xform[3][3], double * SourceVec,
   double * DestVec)
{
   int i,j;

   for (i=0; i<3; i++) {
      DestVec[i] = 0;
      for (j=0; j<3; j++)
         DestVec[i] += xform[i][j] * SourceVec[j];
   }
}


/* this append works for the viewer */

void AppendRotationX(double XformToChange[3][3], double Angle)
{
   double Temp10, Temp11, Temp12, Temp20, Temp21, Temp22;
   double CosTemp = cos(Angle), SinTemp = sin(Angle);

   /* Calculate the new values of the four affected matrix entries */
   Temp10 = CosTemp*XformToChange[1][0]- SinTemp*XformToChange[2][0];
   Temp11 = CosTemp*XformToChange[1][1]- SinTemp*XformToChange[2][1];
   Temp12 = CosTemp*XformToChange[1][2]- SinTemp*XformToChange[2][2];
   Temp20 = SinTemp*XformToChange[1][0]+ CosTemp*XformToChange[2][0];
   Temp21 = SinTemp*XformToChange[1][1]+ CosTemp*XformToChange[2][1];
   Temp22 = SinTemp*XformToChange[1][2]+ CosTemp*XformToChange[2][2];
   /* Put the results back into XformToChange */
   XformToChange[1][0] = Temp10; XformToChange[1][1] = Temp11;
   XformToChange[1][2] = Temp12; XformToChange[2][0] = Temp20;
   XformToChange[2][1] = Temp21; XformToChange[2][2] = Temp22;
}

/* this append works for the object */

/* Concatenate a rotation by angle around the X axis to the transformation in
XformToChange, placing result back in XformToChange. */
void appendx(double xform[3][3], double angle)
{

   double Temp01, Temp11, Temp21, Temp02, Temp12, Temp22;
   double CosTemp = cos(angle), SinTemp = sin(angle);

   /* Calculate the new values of the four affected matrix entries */
   Temp01 = CosTemp*xform[0][1]+ SinTemp*xform[0][2];
   Temp11 = CosTemp*xform[1][1]+ SinTemp*xform[1][2];
   Temp21 = CosTemp*xform[2][1]+ SinTemp*xform[2][2];
   Temp02 = -SinTemp*xform[0][1]+ CosTemp*xform[0][2];
   Temp12 = -SinTemp*xform[1][1]+ CosTemp*xform[1][2];
   Temp22 = -SinTemp*xform[2][1]+ CosTemp*xform[2][2];
   /* Put the results back into xform */
   xform[0][1] = Temp01; xform[1][1] = Temp11;
   xform[2][1] = Temp21; xform[0][2] = Temp02;
   xform[1][2] = Temp12; xform[2][2] = Temp22;

}

/* Concatenate a rotation by Angle around the Y axis to the transformation in
xform, placing result back in xform. */
void appendy(double xform[3][3], double Angle)
{
   double Temp00, Temp10, Temp20, Temp02, Temp12, Temp22;
   double CosTemp = cos(Angle), SinTemp = sin(Angle);

   /* Calculate the new values of the four affected matrix entries */
   Temp00 = CosTemp*xform[0][0]+ -SinTemp*xform[0][2];
   Temp10 = CosTemp*xform[1][0]+ -SinTemp*xform[1][2];
   Temp20 = CosTemp*xform[2][0]+ -SinTemp*xform[2][2];
   Temp02 = SinTemp*xform[0][0]+ CosTemp*xform[0][2];
   Temp12 = SinTemp*xform[1][0]+ CosTemp*xform[1][2];
   Temp22 = SinTemp*xform[2][0]+ CosTemp*xform[2][2];
   /* Put the results back into xform */
   xform[0][0] = Temp00; xform[1][0] = Temp10;
   xform[2][0] = Temp20; xform[0][2] = Temp02;
   xform[1][2] = Temp12; xform[2][2] = Temp22;
}

/* Concatenate a rotation by Angle around the Z axis to the transformation in
xform, placing result back in xform. */
void appendz(double xform[3][3], double Angle)
{
   double Temp00, Temp10, Temp20, Temp01, Temp11, Temp21;
   double CosTemp = cos(Angle), SinTemp = sin(Angle);

   /* Calculate the new values of the four affected matrix entries */
   Temp00 = CosTemp*xform[0][0]+ SinTemp*xform[0][1];
   Temp10 = CosTemp*xform[1][0]+ SinTemp*xform[1][1];
   Temp20 = CosTemp*xform[2][0]+ SinTemp*xform[2][1];
   Temp01 = -SinTemp*xform[0][0]+ CosTemp*xform[0][1];
   Temp11 = -SinTemp*xform[1][0]+ CosTemp*xform[1][1];
   Temp21 = -SinTemp*xform[2][0]+ CosTemp*xform[2][1];
   /* Put the results back into xform */
   xform[0][0] = Temp00; xform[1][0] = Temp10;
   xform[2][0] = Temp20; xform[0][1] = Temp01;
   xform[1][1] = Temp11; xform[2][1] = Temp21;
}


#if 0
/* Concatenate a rotation by angle around the Y axis to the transformation in
XformToChange, placing result back in XformToChange. */
void appendy(double xform[3][3], double angle)
{
double c0, c1, c2;
double length;

/* angle and sin(angle) approx the same for small angles */
/* add as vector */
/* change the direction of Z axis in direction of the X axis */

c0= xform[0][2] + xform[0][0] * angle;
c1= xform[1][2] + xform[1][0] * angle;
c2= xform[2][2] + xform[2][0] * angle;

/* convert back to unit length */
  /* find length of vector */
length= sqrt( c0*c0 + c1*c1 + c2*c2 );

xform[0][2]= c0/length;
xform[1][2]= c1/length;
xform[2][2]= c2/length;

/* make column 0 normal to col 1 and 2 */

xform[0][0]= xform[1][1]*xform[2][2] -
             xform[2][1]*xform[1][2];

xform[1][0]= xform[2][1]*xform[0][2] -
             xform[0][1]*xform[2][2];

xform[2][0]= xform[0][1]*xform[1][2] -
             xform[1][1]*xform[0][2];

}

/* Concatenate a rotation by angle around the Z axis to the transformation in
XformToChange, placing result back in XformToChange. */
void appendz(double xform[3][3], double angle)
{

double a0, a1, a2;
double length;

/* angle and sin(angle) approx the same for small angles */
/* add as vector */
/* change the direction of the X axis in direction of Y axis */

a0= xform[0][0] + xform[0][1] * angle;
a1= xform[1][0] + xform[1][1] * angle;
a2= xform[2][0] + xform[2][1] * angle;

/* convert back to unit length */
  /* find length of vector */
length= sqrt( a0*a0 + a1*a1 + a2*a2 );

xform[0][0]= a0/length;
xform[1][0]= a1/length;
xform[2][0]= a2/length;

/* make column 1 normal to col 2 and 0 */

xform[0][1]= xform[1][2]*xform[2][0] -
             xform[2][2]*xform[1][0];

xform[1][1]= xform[2][2]*xform[0][0] -
             xform[0][2]*xform[2][0];

xform[2][1]= xform[0][2]*xform[1][0] -
             xform[1][2]*xform[0][0];

}
#endif
