/* modified from fftscope.c to show FlyRc tx waveform on parallel port */

#include <stdio.h>
#include <graphics.h>
#include <bios.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include <math.h>
#include <ctype.h>


int samples[512];    /* ADC samples */
int videoy[512];     /* last trace */
int videop[512];     /* pixel value */

int data_port;
int status;
int control;
int invert= 0x40;   /* invert the signal */
int mask= 0x40;     /* mask for ack bit */
long delay_value= 1000;
int holdoff_enable= 1;

void getsamples( int c );

main(){
int x,y;
int c;
int far *p;

   /* get value of lpt1 port */

   p= MK_FP(0,0x408);   /* address of data port is in memory here */
   data_port= *p;
   status = data_port+1;
   control= data_port+2;

   if( setvmode(18) != 18 ){
      setvmode( DEFAULTMODE );
      printf("\n You need a VGA card to run this program");
      exit(1);
      }

   /* draw a grid */

   pen_color(4);
   for( x= 16; x < 516; ++x ){
      setpixel(x,200);
      setpixel(x,115);
      setpixel(x,285);
      setpixel(x,72);
      setpixel(x,328);
      }
   for( x= 16; x < 523; x+= 100 ){
      for( y= 72; y < 329; ++y ){
         setpixel( x,y );
         setpixel( x+1,y );
         }
      }

 /*  for( x= 10; x < 266; ++x ){
      setpixel(x,470);
      } */

   move_to( 530,115 );
   plots(" 5 V");
   move_to( 530,285 );
   plots("-5 V");
   move_to( 430,50 );
   plots("5ms/div");

   /* run continuous plots */

   outportb(control,0x4); /* set our pull up resistor high */

   while ( 1 ){
      if( holdoff_enable ) holdoff();
      sync();
      getsamples( 512 );  /* get and plot */
      if( kbhit() ){
         pen_color(15);
         c= tolower(getch());
         if( c == 27 ) break;   /* exit on escape */
         if( c == 'i' ) invert^= 0x40;  /* toggle invert */
         if( c == 'h' ) holdoff_enable ^= 1;
         }
      }

   setvmode( DEFAULTMODE );
   printf(" %x %ld %x\n",data_port,delay_value,invert);

}



void getsamples( int number ){   /* get the number of samples */
int x,y;                           /* plot them here also */
static int first_time= 1;
int signal;

   for( x= 0; x < number; ++x ){    /* fill the sample array */
      samples[x]= recv();
      }

   signal= 0;
   for( x= 0; x < number; ++x ){
      if( first_time == 0 ){
         y= videoy[x];
         pen_color(videop[x]);
         setpixel(x+10,y);
         }
      if( samples[x] != samples[0] ) signal= 1;
      pen_color(15);
      if( samples[x] < 128 ) y= 200 - samples[x];    /* pos */
      else y= 200 - samples[x] + 256;                /* neg */
      videoy[x]= y;
      videop[x]= getpixel(x+10,y);
      setpixel(x+10,y);
      }

   first_time= 0;

   /* adjust delay values */
   if( signal ){
      for( x= 0; x < 10; ++x ) if( samples[x] != 0 ) --delay_value;
      if( samples[10] == 0 ) ++delay_value;
      }

}


int recv(){    /* receive value on comport */
int counter;
int c;
int my_delay;

   my_delay= (int) delay_value;
   counter= 0;
   while( counter++ < my_delay );  /* delay */

   c=  inportb(status);
   c= c & mask;
   c= c ^ invert;
   return c;

}


holdoff(){   /* o'scope holdoff - wait for stable signal */
int c;
int old;
int count;
int counter2;

   old= c= recv();
   counter2= 0;
   while( count < 100 ){    /* get 100 samples all alike */
      c= recv();
      if( c != old ){
         count= 0;
         old= c;
         }
      if( counter2++ > 30000 )   break;
      }
}

sync(){      /* sync on neg going signal */
long counter;
int c;
long my_delay;

   counter= 0;
   my_delay= delay_value * 1000;

   while( counter++ < my_delay ){
      c= inportb(status);
      c = c & mask;
      c= c ^ invert;
      if( c ) break;
      }

   counter= 0;
   while( counter++ < my_delay ){
      c= inportb(status);
      c= c & mask;
      c= c ^ invert;
      if( c == 0 ) break;
      }
}


